module Tutorial7 where

import LSystem
import Test.QuickCheck

pathExample = Go 30 :#: Turn 120 :#: Go 30 :#: Turn 120 :#: Go 30

-- 1a. copy
copy :: Int -> Command -> Command
copy 0 _ = Sit
copy n c = c :#: copy (n-1) c

-- 1b. polygon
polygon :: Distance -> Int -> Command
polygon s n = copy n (Go s :#: Turn (360.0/n'))
                where n' = fromIntegral n

-- 2. snowflake
snowflake :: Int -> Command
snowflake x = GrabPen blue :#: f x :#: n :#: n :#: f x :#: n :#: n :#: f x :#: n :#: n
  where
  f 0 = Go 10 
  f x = f(x-1) :#: p :#: f(x-1) :#: n :#: n :#: f(x-1) :#: p :#: f(x-1)
  p = Turn 60
  n = Turn (-60)
           

-- 3. sierpinski
sierpinski :: Int -> Command
sierpinski = f
  where
  f 0 = GrabPen red :#: Go 10
  f x = g (x-1) :#: p :#: f (x-1) :#: p :#: g (x-1)
  g 0 = GrabPen blue :#: Go 10
  g x = f (x-1) :#: n :#: g (x-1) :#: n :#: f (x-1)
  p = Turn 60
  n = Turn (-60)
     
-- 4. hilbert
hilbert :: Int -> Command
hilbert =  l
  where
  l 0 = Sit
  l x = p :#: r (x-1) :#: f :#: n :#: l (x-1) :#: f :#: l (x-1) :#: n :#: f :#: r (x-1) :#: p
  r 0 = Sit
  r x = n :#: l (x-1) :#: f :#: p :#: r (x-1) :#: f :#: r (x-1) :#: p :#: f :#: l (x-1) :#: n
  f  = Go 10
  p = Turn 90
  n = Turn (-90)

-- 5. dragon
dragon :: Int -> Command
dragon =  l
  where
  l 0 = Sit
  l x = l (x-1) :#: p :#: r (x-1) :#: f :#: p
  r 0 = Sit
  r x = n :#: f :#: l (x-1) :#: n :#: r (x-1)
  f = Go 10
  p = Turn 90
  n = Turn (-90)

-- ** Optional Material

-- 6a. split
split :: Command -> [Command]
split (c :#: c') = split c ++ split c'
split Sit = []
split c = [c]

-- 6b. join
join :: [Command] -> Command
join = foldr (:#:) Sit

-- 6c. equivalent
equivalent :: Command -> Command -> Bool
equivalent c c' = split c == split c'

-- 6d. testing join and split
prop_split_join :: Command -> Bool
prop_split_join c =  equivalent (join $ split c) c

prop_split :: Command -> Bool
prop_split c = and [ isLiteral c' | c' <- split c ]
                 where
                 isLiteral :: Command -> Bool
                 isLiteral (c :#: c') = False
                 isLiteral Sit = False
                 isLiteral _ = True

-- 7. optimise
join' :: [Command] -> Command
join' = foldr1 (:#:)

optimise :: Command -> Command
optimise c = join' (combine (split c))
               where
               combine :: [Command] -> [Command]
               combine [] = []
               combine [Go 0] = []
               combine [Turn 0] = []
               combine [Go x] = [Go x]
               combine [Turn x] = [Turn x]
               combine (Go x : Go y : c) = combine (combine [Go (x+y)] ++ combine c)
               combine (Turn x : Turn y : c) = combine (combine [Turn (x+y)] ++ combine c)
               combine (c : c') = combine [c] ++ combine c'

challenge :: Int -> Command
challenge = f
  where
  f 0 = GrabPen red :#: Go 10
  f x = g (x-1) :#: p' :#: i (x-1) :#: Branch (p :#: k (x-1)) :#: n' :#: f (x-1) :#: p :#: h (x-1)
  g 0 = GrabPen blue :#: Go 10
  g x = h (x-1) :#: n :#: f (x-1) :#: n :#: f (x-1) :#: n' :#: g (x-1)
  h 0 = GrabPen green :#: Go 10
  h x = h (x-1) :#: p :#: g (x-1) :#: n :#: i (x-1) :#: Branch (p :#: f (x-1))
  i 0 = GrabPen black :#: Go 10
  i x = f (x-1) :#: p :#: p :#: k (x-1) :#: n' :#: g (x-1) :#: n :#: n :#: h(x-1) :#: p' :#: i (x-1)
  k 0 = GrabPen white :#: Go 10
  k x = i (x-1) :#: Branch (n' :#: g (x-1)) :#: f (x-1) :#: p' :#: k (x-1)
  p = Turn 10
  n = Turn (-10)
  p' = Turn 40
  n' = Turn (-40)
